# -*- coding: utf-8 -*-
from __future__ import absolute_import

__version__ = '0.3.0'
__author__ = 'Steven Loria'
__license__ = "MIT"

from marshmallow.serializer import Serializer

from marshmallow.utils import pprint
